#include "player_character.hpp"
#include <godot_cpp/classes/navigation_agent3d.hpp>
#include "godot_cpp/variant/utility_functions.hpp"
#include "utils/godot_macros.h"

namespace godot {
void PlayerCharacter::_bind_methods() {
#define CLASSNAME PlayerCharacter
}

void PlayerCharacter::_enter_tree() { GDGAMEONLY();
    this->nav_agent = this->get_node<NavigationAgent3D>("NavigationAgent3D");
    this->target_rotation = this->get_global_transform().get_basis().get_quaternion();
}

void PlayerCharacter::_process(double delta_time) { GDGAMEONLY();
    if(!this->mode_manual) {
        this->process_ai(delta_time);
    }
}

void PlayerCharacter::_physics_process(double delta_time) { GDGAMEONLY();
    Vector3 const new_velocity = this->get_velocity().move_toward(this->velocity_target, delta_time * PlayerCharacter::ACCELERATION);
    UtilityFunctions::print("Velocity target: ", this->velocity_target);
    UtilityFunctions::print("Velocity current: ", this->get_velocity());
    UtilityFunctions::print("Velocity next: ", new_velocity);
    this->set_velocity(new_velocity);
    this->move_and_slide();
}

void PlayerCharacter::move(Vector3 world_vector) {
    this->velocity_target = world_vector * PlayerCharacter::SPEED;
}

void PlayerCharacter::aim(Vector3 at) {
    Vector3 const position{this->get_global_position()};
    Vector3 const forward{(Vector3{at.x, 0.f, at.z} - Vector3{position.x, 0.f, position.z}).normalized()};
    Vector3 const up{0.f, 1.f, 0.f};
    this->target_rotation = {forward.cross(up), up, forward};
}

void PlayerCharacter::set_manual_mode(bool value) {
    this->mode_manual = value;
}

void PlayerCharacter::process_ai(double delta_time) {
    this->velocity_target = this->nav_agent->get_velocity();
}

void PlayerCharacter::process_rotation(double delta_time) {
    Transform3D current_trans{this->get_global_transform()};
    Basis const current_basis{current_trans.get_basis()};
    real_t const angle{current_basis.get_quaternion().angle_to(this->target_rotation.get_quaternion())};
    current_trans.set_basis(current_basis.lerp(this->target_rotation, PlayerCharacter::ROTATION_SPEED / angle * delta_time));
    this->set_global_transform(current_trans);
}

float const PlayerCharacter::ACCELERATION{100.f};
float const PlayerCharacter::SPEED{3.f};
float const PlayerCharacter::ROTATION_SPEED{1.f};
}
