#include "tunnels_player.hpp"
#include "godot_cpp/classes/resource_loader.hpp"
#include "player_character.hpp"
#include "utils/game_root.hpp"
#include "utils/godot_macros.h"
#include "utils/player_input.hpp"
#include <godot_cpp/classes/input_event.hpp>
#include <godot_cpp/classes/scene_state.hpp>
#include <godot_cpp/classes/viewport.hpp>

namespace godot {
void TunnelsPlayer::_bind_methods() {
#define CLASSNAME TunnelsPlayer
    GDFUNCTION_ARGS(horizontal_move_input, "event", "value");
    GDFUNCTION_ARGS(vertical_move_input, "event", "value");
}

void TunnelsPlayer::_ready() { GDGAMEONLY();
    this->camera = this->get_viewport()->get_camera_3d();
    this->initialize_character();
}

void TunnelsPlayer::_exit_tree() { GDGAMEONLY();
    GameRoot::get_singleton()->remove_player(this->get_player_id());
}

void TunnelsPlayer::_process(double delta_time) { GDGAMEONLY();
    switch(this->state) {
    default:
    case State::ManualControl:
        this->character->move(this->get_world_move_input().normalized());
        this->set_global_position(this->character->get_global_position());
        break;
    case State::Tactics:
        break;
    case State::Overview:
        this->set_global_position(this->get_global_position() + this->get_world_move_input().normalized());
        break;
    }
}

void TunnelsPlayer::setup_player_input(PlayerInput *input) {
    input->listen_to(PlayerInput::Listener("move_right", "move_left", this, "horizontal_move_input"));
    input->listen_to(PlayerInput::Listener("move_forward", "move_backward", this, "vertical_move_input"));
}

Node *TunnelsPlayer::to_node() {
    return Object::cast_to<Node>(this);
}

void TunnelsPlayer::horizontal_move_input(Ref<InputEvent> event, float value) {
    this->move_input.x = value;
}

void TunnelsPlayer::vertical_move_input(Ref<InputEvent> event, float value) {
    this->move_input.y = value;
}

void TunnelsPlayer::initialize_character() {
    Ref<PackedScene> player_scene = ResourceLoader::get_singleton()->load("res://player_character.tscn");
    if(player_scene.is_null() || !player_scene.is_valid())
        return;
    if(player_scene->get_state()->get_node_type(0) != StringName("PlayerCharacter"))
        return;
    this->character = Object::cast_to<PlayerCharacter>(player_scene->instantiate());
    this->get_parent()->add_child(this->character);
    this->character->set_global_transform(this->get_global_transform());
    this->character->set_manual_mode(true);
}

Vector3 TunnelsPlayer::get_world_move_input() const {
    Basis const camera_basis = camera->get_global_transform().get_basis();
    // get the forward and left vectors, ensuring that they won't produce {0,0,0} when flattened
    Vector3 x = camera_basis.get_column(0);
    if(x.x == 0.f && x.z == 0.f) x = camera_basis.get_column(1);
    Vector3 z = camera_basis.get_column(2);
    if(z.x == 0.f && z.z == 0.f) z = camera_basis.get_column(1);
    return this->move_input.x * Vector3{x.x, 0.f, x.z}.normalized()
         + this->move_input.y * Vector3{z.x, 0.f, z.y}.normalized();
}
}
