#ifndef TUNNELS_PLAYER_HPP
#define TUNNELS_PLAYER_HPP

#include "utils/player.hpp"
#include "utils/player_input.hpp"
#include <godot_cpp/classes/camera3d.hpp>
#include <godot_cpp/classes/character_body3d.hpp>
#include <godot_cpp/classes/input_event.hpp>

namespace godot {
class PlayerCharacter;

class TunnelsPlayer : public Node3D, public IPlayer {
    GDCLASS(TunnelsPlayer, Node3D);
    static void _bind_methods();

    enum State {
        ManualControl = 0x0,
        Tactics = 0x1,
        Overview = 0x2,
    };

public:
    virtual void _ready() override;
    virtual void _exit_tree() override;
    virtual void _process(double delta_time) override;

    virtual void setup_player_input(PlayerInput *input) override;
    virtual Node *to_node() override;

    void horizontal_move_input(Ref<InputEvent> event, float value);
    void vertical_move_input(Ref<InputEvent> event, float value);

    void initialize_character();

    Vector3 get_world_move_input() const;
private:
    Vector2 move_input{0,0};
    Vector2 mouse_location{0,0};
    PlayerCharacter *character{nullptr};
    TunnelsPlayer::State state{State::ManualControl};

    Camera3D *camera{nullptr};
};
}

#endif // !TUNNELS_PLAYER_HPP
